"""
Armor MCP Package

This package provides an agentic interface for interoperating with multiple blockchains,
staking, DeFi operations, swaps, bridging, wallet management, and for developing crypto
trading strategies through dynamic DCA.
"""

__version__ = "0.1.23"


